#!/usr/bin/env python3

#   gtemp.py: Calculate average temperature anomaly from GHCN-M data.
#   Copyright (C) 2017  Climate Change Truth Inc.
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import math
import subprocess

months = ["January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"]

# First argument is filename.
parser = argparse.ArgumentParser(description="Calculate global temperature average from GHCN-M data")
parser.add_argument("stations", metavar="STATIONS", type=argparse.FileType("r"),
	help="Station data")
parser.add_argument("data", metavar="MEASUREMENTS", type=argparse.FileType("r"),
	help="Measurement data")
parser.add_argument("-a", "--after", metavar="YEAR", type=int,
	help="Only include data after and including (>=) YEAR")
parser.add_argument("-b", "--before", metavar="YEAR", type=int,
	help="Only include data before (<) YEAR")
parser.add_argument("-o", "--output", metavar="NAME", type=str, default="output",
	help="Output file name (without '.png' extension)")
parser.add_argument("-t", "--title", metavar="TITLE", type=str,
	help="Override automatic graph title generation with TITLE")
parser.add_argument("--north-from", metavar="DEG", type=float,
	help="Only include latitudes from DEG and northward")
parser.add_argument("--south-from", metavar="DEG", type=float,
	help="Only include latitudes from DEG and southward")
parser.add_argument("--yrange-max", metavar="DEG", type=float, default=2.5,
	help="Y-range minimum for the graph")
parser.add_argument("--yrange-min", metavar="DEG", type=float, default=-1,
	help="Y-range minimum for the graph")
args = parser.parse_args()

# Read in data.
#   - $station
#     - "year"
#       - $year
#         - $month
#           - "value"
#           - "dmflag"
#           - "qcflag"
#           - "dsflag"
lines = list(line.strip("\n") for line in args.data.readlines())
data = {}
for line in lines:
	station = line[0:11]
	year = line[11:15]
	element = line[15:19]

	if (args.after and int(year) < args.after):
		#print("Data from '%s' is not at or after '%s'; skipping" % (year, args.after))
		continue
	elif (args.before and int(year) >= args.before):
		#print("Data from '%s' is not before '%s'; skipping" % (year, args.before))
		continue

	# Check for unknown element tag.
	if element != "TAVG":
		print("Unexpected element '%s'; expected 'TAVG', skipping" % (element))
		continue

	if station not in data.keys():
		#print("Added station: '%s'", station)
		data[station] = {"years" : {}}
	if year in data[station]["years"].keys():
		print("Station '%s' contains duplicate year '%s'; skipping duplicate" % (station, year))
		continue
	else:
		data[station]["years"][year] = {}
	for i, month in zip(range(19, 115, 8), months):
		value = line[i:i+5]
		dmflag = line[i+5]
		qcflag = line[i+6]
		dsflag = line[i+7]

		data[station]["years"][year][month] = {}
		data[station]["years"][year][month]["value"] = value
		data[station]["years"][year][month]["dmflag"] = dmflag
		data[station]["years"][year][month]["qcflag"] = qcflag
		data[station]["years"][year][month]["dsflag"] = dsflag

# Compute station averages for each month over the 1961-1990 base period
#   - $station
#     - "average"
#       - $month
#         - value
for station in data.keys():
	data[station]["average"] = {}

	# Compute station average for each month
	for month in months:
		data[station]["average"][month] = None
		values = []
		# Compute station average for given month
		for year in data[station]["years"].keys():
			if int(year) < 1961 or int(year) > 1990:
				continue # Base period only from 1961-1990
			#print("%s %s %s %s" % (station, year, month, data[station]["years"][year][month]))
			value = data[station]["years"][year][month]["value"]
			if value == "-9999":
				continue # Missing data.
			values.append(int(value))

		if len(values) < 20:
			continue # Not enough data for given month.
		data[station]["average"][month] = float(sum(values)) / len(values)
		#print("Station '%s', month '%s', average '%s'" % (station, month, data[station]["average"][month]))

# Read in station metadata.
#   - $station
#     - "latitude"
#     - "longitude"
lines = list(line.strip("\n") for line in args.stations.readlines())
for line in lines:
	station = line[0:11]
	lat = line[12:20]
	lon = line[21:30]

	if station not in data.keys():
		print("Station '%s' found in station metadata but no data for station, skipping" % station)
		continue
	data[station]["latitude"] = lat
	data[station]["longitude"] = lon

	# Late lat/lon checks.
	if (args.north_from is not None and float(lat) < args.north_from):
		print("Removing station '%s' at lat '%s'" % (station, lat))
		del data[station]
	if (args.south_from is not None and float(lat) > args.south_from):
		print("Removing station '%s' at lat '%s'" % (station, lat))
		del data[station]

# Create gboxes.
gboxes = {}
for lat in range(0, 180, 5):
	gboxes[lat] = {}
	for lon in range(0, 360, 5):
		gboxes[lat][lon] = {"stations": {}, "averages": {}}

# Calculate station anomalies in 5x5 grid boxes.
#   - "gboxes"
#     - $lat
#       - $long
#         - "stations"
#           - $station
#             - $year
#               - $month
for station in data.keys():
	# Calculate appropriate gbox.
	lat = (int(float(data[station]["latitude"].strip())) + 90)
	lat = lat - (lat % 5)
	lon = (int(float(data[station]["longitude"].strip())) + 180)
	lon = lon - (lon % 5)

	# Add station data to gbox.
	for year in data[station]["years"].keys():
		for month in data[station]["years"][year].keys():
			if not data[station]["average"][month]:
				continue  # Not enough data for station average temp for this month.
			elif data[station]["years"][year][month]["value"] == "-9999":
				continue  # Missing data.

			if station not in gboxes[lat][lon]["stations"].keys():
				gboxes[lat][lon]["stations"][station] = {}
			if year not in gboxes[lat][lon]["stations"][station].keys():
				gboxes[lat][lon]["stations"][station][year] = {}
			gboxes[lat][lon]["stations"][station][year][month] = float(data[station]["years"][year][month]["value"]) - float(data[station]["average"][month])

# Average station anomalies per grid box.
#   - "gboxes"
#     - $lat
#       - $long
#         - "average"
#           - $year
#             - $month
for lat in gboxes.keys():
	for lon in gboxes[lat].keys():
		# Gather data from each station.
		for station in gboxes[lat][lon]["stations"].keys():
			for year in gboxes[lat][lon]["stations"][station].keys():
				for month in gboxes[lat][lon]["stations"][station][year].keys():
					if year not in gboxes[lat][lon]["averages"].keys():
						gboxes[lat][lon]["averages"][year] = {}
					if month not in gboxes[lat][lon]["averages"][year].keys():
						gboxes[lat][lon]["averages"][year][month] = []
					gboxes[lat][lon]["averages"][year][month].append(gboxes[lat][lon]["stations"][station][year][month])
		# Average data.
		for year in gboxes[lat][lon]["averages"].keys():
			for month in gboxes[lat][lon]["averages"][year].keys():
				gboxes[lat][lon]["averages"][year][month] = sum(gboxes[lat][lon]["averages"][year][month]) / len(gboxes[lat][lon]["averages"][year][month])
				#print("Gbox average: lat '%s', lon '%s', year '%s', month '%s', average: '%s'" % (lat, lon, year, month, gboxes[lat][lon]["averages"][year][month]))

# Weighted average of all grid boxes
averages = {}
for lat in gboxes.keys():
	for lon in gboxes[lat].keys():
		# Gather all year+month averages for each gbox.
		for year in gboxes[lat][lon]["averages"].keys():
			for month in gboxes[lat][lon]["averages"][year].keys():
				if year not in averages.keys():
					averages[year] = {}
				if month not in averages[year].keys():
					averages[year][month] = []
				averages[year][month].append((gboxes[lat][lon]["averages"][year][month], math.cos(math.radians(lat - 90 + 2.5))))
# Average each gbox.
for year in averages.keys():
	for month in averages[year].keys():
		averages[year][month] = sum(data[0] * data[1] for data in averages[year][month]) / sum(data[1] for data in averages[year][month])

# Weighted average per year.
for year in averages.keys():
	averages[year] = sum(averages[year][month] for month in averages[year].keys()) / len(averages[year].keys())
	averages[year] = averages[year] / 100

# Fill in empty years.
years = sorted(averages.keys())
for year in range(int(years[0]), int(years[-1])):
	year = str(year)
	if year not in averages.keys():
		averages[year] = ""

# Write data.
file = open(".temp.dat", "w")
for year in sorted(averages.keys()):
	label = year
	if int(year) % 10:
		label = ""
	else:
		label = year
	file.write("%s,%s\n" % (label, averages[year]))
file.close()

# Generate plot title.
title = "GHCNM-v3.3.0 "
params = []
if args.north_from is not None:
	params.append("North From %s\u00b0 Latitude" % args.north_from)
if args.south_from is not None:
	params.append("South From %s\u00b0 Latitude" % args.south_from)
if not params:
	title += "Global "
else:
	title += ", ".join(params) + " "
title += "Average Temperature Anomaly"
if args.title is not None:
	title = args.title

# Plot data.
gnuplot = subprocess.Popen(['gnuplot'], stdin=subprocess.PIPE)
gnuplot.communicate((
	"stats '.temp.dat' using 1 name 'DATA'\n" +
        "stats '.temp.dat' using 0 name 'DATA2'\n" +
	("set title '%s'\n" % title) +
	"set style data linespoints\n"
	"set grid\n"
	"set datafile separator ','\n"
	"set terminal png size (100+DATA_records*170),600\n" +
	("set output '%s.png'\n" % args.output) +
	"set nokey\n"
	"set xtics rotate by 45 right\n"
	#"set xtics font ', 10'\n"
	"set xlabel 'Year'\n" +
	#"set ytics -1, 1, 2\n"
	("set yrange [%s:%s]\n" % (args.yrange_min, args.yrange_max)) +
	"set ylabel 'Average Temperature Anomaly (in degrees Celcius)'\n"
        "last_index(i) = (i == DATA2_records - 1 ? 1 : 0)\n"
        "plot '.temp.dat' using 2:xtic(1) title 'Temperature', '' u 0:2:(last_index($0) ? sprintf(\"%.2f\", $2) : '') w labels right offset -0.3\n"
).encode())
